/* 
   (c) COPYRIGHT URI/MIT 1996,1997,1999
   Please read the full copyright statement in the file COPYRIGHT.

   Authors:
      grf		Glenn Fleirl (glenn@lake.mit.edu)
      jhrg,jimg		James Gallagher (jgallagher@gso.uri.edu)
*/

/* 
   loaddods is a Matlab 4 command which loads the variables specified by a
   DODS URL into Matlab. Because of limitations on the data types which
   Matlab 4 can process, there are some DODS datasets, or parts of some
   datasets, which cannot be read. Because the DODS Sequence data type is
   used by many important datasets, this software translates that type to an
   array or collection of arrays.

   jhrg 10/6/96 
*/

/* 
 * $Log: loaddods.c,v $
 * Revision 1.40  1999/07/24 00:10:29  jimg
 * Merged the release-3-0-2 branch
 *
 * Revision 1.39  1999/04/30 17:06:57  jimg
 * Merged with no-gnu and release-2-24
 *
 * Revision 1.38.2.2  1999/04/29 20:13:00  jimg
 * Removed old code.
 *
 * Revision 1.38.2.1  1999/04/29 20:07:50  jimg
 * Fixed error messages so that they either print text to stderr *or* store
 * text in a matlab variable called dods_err_msg. Use the -w flag to get this
 * behavior.
 * Changed the -V option so that it returns the version info to a matrix when
 * loaddods is called using the assignment syntax.
 *
 * Revision 1.38  1999/03/29 21:25:36  jimg
 * Added support for the new datatypes
 *
 * Revision 1.37  1999/02/08 20:55:00  jimg
 * Fixed a bug where PATH values longer than 256 chars were not searched. This
 * meant that if users put writeval in a directory that was listed at the end of
 * their path and strlen(PATH) was > 256 chars, writeval would not be found.
 * However, it *would* be found if listed at the front of PATH. I made the
 * software more general and, at the same time, added search of MATLABPATH. That
 * is, now DODS_ROOT/{etc,bin}, PATH, MATLABPATH and the CWD are all searched
 * when loaddods looks for writeval.
 *
 * Revision 1.36  1998/11/30 05:37:36  jimg
 * Added the -T option to control truncation of names longer than 19 characters
 *
 * Revision 1.35  1998/11/25 01:15:23  jimg
 * Identifiers limited to 19 chars. This is a temp fix; will update to include ML 5 later
 *
 * Revision 1.34  1998/11/20 08:41:22  jimg
 * Fixed the fix for the extra newline separators; it was broken for arrays
 *
 * Revision 1.33  1998/11/19 20:54:37  jimg
 * Repaired some error messages
 *
 * Revision 1.32  1998/11/10 03:18:14  jimg
 * Added test stuff. This is fairly hokey...
 *
 * Revision 1.31  1998/06/09 04:04:36  jimg
 * Fixed get_pathname.
 * Added exit(1) to do_error().
 *
 * Revision 1.30  1998/03/21 00:20:54  jimg
 * Fixed string internment so that it now works with vectors of strings (using
 * reorganize_strings()).
 *
 * Revision 1.29  1998/03/19 23:24:36  jimg
 * Added do_error(). It reads error messages relayed by writeval and displays
 * them using mexErrMsgTxt().
 *
 * Revision 1.28  1997/12/16 02:58:06  jimg
 * Merged release 2.14d changes
 *
 * Revision 1.27  1997/10/24 23:52:43  jimg
 * Added to get_pathname() so that the CWD is searched first.
 *
 * Revision 1.26  1997/10/09 22:20:05  jimg
 * Resolved conflicts in merge of 2.14c to trunk.
 *
 * Revision 1.25  1997/10/04 00:33:48  jimg
 * Release 2.14c fixes
 *
 * Revision 1.23.6.1  1997/09/19 19:11:22  jimg
 * Moved config_writeval.h so that -V4_COMPAT flag works.
 *
 * Revision 1.24  1997/09/10 22:49:56  jimg
 * Fixed a bug that shows up on Solaris 2 where the pipe connecting this
 * program to writeval must be flushed before calling pclose().
 *
 * Revision 1.23  1997/05/06 00:34:26  jimg
 * Updated version
 *
 * Revision 1.22  1997/05/01 22:40:59  jimg
 * loaddods can now return arguments using Matlab's assignment operator:
 * [a b c] = loaddods('...'); In addition, it now accepts the option -s which
 * causes the first return argument to be a vector of strings (elements
 * separated by a tab) naming the variables in the following return arguments.
 *
 * Revision 1.21  1997/04/29 23:42:39  jimg
 * Fixed a bug where loaddods crashed Matlab when invoked with '-v' and no
 * other options. This also fixes the bug where command line options broke
 * reading from the urlqueue program.
 *
 * Revision 1.20  1997/04/21 20:29:23  jimg
 * Updated version to 1.4 - this matches version 2.13a of DODS.
 *
 * Revision 1.19  1997/04/21 18:29:10  jimg
 * get_pathname() now performs tilde (~) expansion ala csh(1).
 *
 * Revision 1.18  1997/04/19 00:49:01  jimg
 * Now checks the user's PATH and DODS_ROOT/etc before calling popen(). popen()
 * does not return an error if the shell (sh) runs but cannot find the program.
 *
 * Revision 1.17  1997/04/18 22:45:53  jimg
 * Added #include <sys/msg.h> for lunix.
 * Removed sys/wait.h since I now use popen/pclose.
 *
 * Revision 1.16  1997/04/18 21:12:26  jimg
 * Fixed a bug which (initially) showed up only on Solaris (but later on
 * SunOS too) where Grids whose size was a power of 2 failed to load. The
 * problem was caused by using fseek() to read a newline character that
 * followed the data. I the course of finding this bug I removed all use
 * of fscanf() and isolate reads from the writeval subprocess into a
 * handful of read_*() functions. Also added was a `test mode' where
 * loaddods could be built as a regular UNIX executable (to simplify
 * debugging).
 *
 * Revision 1.15  1997/03/27 22:57:59  jimg
 * Added code on do_array to ensure that only column-major data is interned in
 * Matlab. This currently entails a stop-and-copy type operation. A more
 * thorough fix would transform the row-major data from writeval while reading
 * from the pipe...
 *
 * Revision 1.14  1997/03/27 18:34:02  jimg
 * Fixed quoting bug - double quotes were not escaped with backslash.
 *
 * Revision 1.13  1997/03/23 22:33:15  jimg
 * Changed so that concatenation is off by default. Use `-k' to activate it.
 * Added `-n' option for the fullnames feature.
 *
 * Revision 1.12 1997/03/07 04:47:04 jimg Added catenation option. Variables
 * are by default catenated when they have the same name. Note that this only
 * works for discrete calls to loaddods. That is, if you call loaddods(...)
 * and it contains several URLs which read variables with the same name,
 * those vectors or arrays will be catenated into single variables. Use the
 * `+k' option to suppress this.
 *
 * Revision 1.11  1997/02/11 16:44:26  jimg
 * Fixed a bug in quote_for_shell() where trailing non-quoted chars were
 * truncated.
 *
 * Revision 1.10  1997/02/06 19:38:42  jimg
 * Added constants for string sizes.
 * Fixed the `absolute path' bug where writeval would only be run if it was in
 * the user's CWD.
 * Added recognition of NaN; when writeval emits NaN for a Float64 value,
 * loaddods uses Matlab's mexGetNaN to intern the value.
 * Verbose mode is now more informative.
 *
 * Revision 1.9  1997/02/03 20:19:53  jimg
 * Removed startchild() from loaddods and used popen instead. The code that
 * checks for the processes started by startchild() does not work correctly,
 * resulting in a `no more pipes' error message when loaddods is used
 * repeatedly. popen()/pclose() fix that problem.
 * Also, removed use of exit() since it causes the matlab process itself to
 * quit.
 *
 * Revision 1.8  1997/01/08 21:32:47  jimg
 * Added UInt32 to set of scalar types processed by do_array(). This fixes a
 * problem where arrays of UInt32 were not processed properly.
 * Added UInt32 to the set of type names that loaddods will process using
 * do_float().
 *
 * Revision 1.7  1996/12/21 00:42:54  jimg
 * Fixed a nasty bug where single dimension arrays resulted in HUGE chunks of
 * memory being allocated. See do_array() and its use of mexPutFull().
 *
 * Revision 1.6  1996/12/18 00:50:34  jimg
 * Added the ability to process command switches which are passed to writeval.
 * loaddods can be invoked `loaddods('-g', '*')' to get the DODS list in
 * netscape and have the GUI displayed by writeval.
 *
 * Revision 1.5  1996/10/29 21:40:34  jimg
 * Added fullnames global that controls whether interned variables are named
 * based on their position in the DDS hierarchy (fullnames == true) or based on
 * their `leaf name' only (fillnames == false). The default is false since when
 * true names can become too long for Matlab.
 * Added error checking for some more of the mexPutFull calls.
 * Replaced fscanf reads of newline with fseek() calls since on the Dec Alpha
 * the hex sequence 0a0c (newline followed by 12(dec)) is read for "\n" by
 * fscanf() when this software expects that only the 0x0a byte will be read.
 * This bug may not be limited to the Dec Alpha, however, since that machine is
 * little endian the 0x0a0c pattern showed up frequently.
 *
 * Revision 1.4  1996/10/25 22:59:52  jimg
 * Added default locator as constant.
 * Added debugging statements.
 * Fixed problems with do_grid.
 * Added verbose flag - there is currently no way to control this option, it is
 * always on.
 * Removed do_function - use do_sequence instead.
 *
 * Revision 1.3  1996/10/23 23:58:39  jimg
 * Massive rewrite so that recursive data types are not flattened by instead
 * are printed `recursively'.
 *
 * Revision 1.2  1996/10/07 20:56:11  jimg
 * Broke out the queue functions to their own file (which is also used by
 * urlqueue).
 * Added header. */

#include "config_writeval.h"

static char id[]={"$Id: loaddods.c,v 1.40 1999/07/24 00:10:29 jimg Exp $"};

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

#include <sys/ipc.h>
#include <sys/msg.h>

#include "mex.h"

#include "queue.h"
#include "variable.h"
#include "defines.h"

#ifdef TEST_LOADDODS
/*
  We *should* be using <nan.h> to handle generating NaN. However, many
  machines don't have a function that can be used to generate NaN so I've
  written my own (which uses its own bastardized version of the FP union).
  2/5/97 jhrg
*/

#include "nan_hack.h"

static double
MakeNaN() 
{ 
    union dnan X;
    X.nan.exponent = 0x7ff; 
    X.nan.qnan_bit = 0x1; 
    
    return X.d;
}

#endif /* TEST_LOADDODS */

#define MATLAB_COMMAND "load %s"
#define BROWSER_COMMAND "netscape -remote 'openURL(%s)'"
#define DODS_BASE_COMMAND "writeval"
#define DODS_COMPLETE_COMMAND "%s -f %s -- %s"
#define DEFAULT_LOCATOR "http://unidata.ucar.edu/packages/dods/"
#define MAX_STR 256
#define BIG_STR 2048

#ifdef DODS_VERSION
static char *dods_version = DODS_VERSION;
#else
static char *dods_version = "unknown";
#endif /* VERSION */

static int verbose = 1;
static int extend_existing_variables = 0;
static int fullnames = 0;
static char command_opts[MAX_STR] = ""; /* Options passed to writeval */
static int output_variable_names = 0;

/** If NUM_RETURN_ARGS is false, use the default mechanism where variables are
   directory interned in Matlab. If true, return values via plhs[]. If true
   the value of NUM_RETURN_ARGS is the number of values to return. After
   reading this many variables, discard any remaining ones. */
int num_return_args = 0;
/** Global array of arrays used to hold NUM_RETURN_ARGS. */
Matrix **return_args;
/** The major version number of Matlab. */
int version_major = 0;
/** Truncate the variable names at 19 characters? 
    @see extend.c */
int truncate_variable_names = FALSE;

extern int intern(char *name, int r, int c, double *yp, int extend);
extern void clear_existing_variables();
extern char *get_variable_name(int first);

extern int current_arg;		/* index of current argument number */

int do_error(FILE *, variable **vectors, char *, int);
int do_float(FILE *, variable **vectors, char *, int);
int do_string(FILE *, variable **vectors, char *, int);
int do_list(FILE *, variable **vectors, char *, int);
int do_array(FILE *, variable **vectors, char *, int);
int do_structure(FILE *, variable **vectors, char *, int);
int do_sequence(FILE *, variable **vectors, char *, int);
int do_grid(FILE *, variable **vectors, char *, int);

typedef (*reader_type)(FILE *, variable **vectors, char *, int);

/* NB: The Byte and Int32 types are mapped to Float64 since Matlab 4
   represents numbers using only 64 bit floating point values. */

struct typeitem {
    char typename[MAX_STR];
    reader_type do_type;
} typelist[] = {
    {"Error", do_error},
    {"Byte", do_float},
    {"Int16", do_float},
    {"UInt16", do_float},
    {"Int32", do_float},
    {"UInt32", do_float},
    {"Float32", do_float},
    {"Float64", do_float},
    {"String", do_string},
    {"Url", do_string},
    {"List", do_list},
    {"Array", do_array},
    {"Structure", do_structure},
    {"Sequence", do_sequence},
    {"Function", do_sequence},
    {"Grid", do_grid},
    {"", NULL}
};

/** Strip out certain characters from a string.

    @param str Operate on this char *.
    @param c Look for this character and remove it. */
static void
strip(char *dest, char *src, char c)
{
    while (*src) {
	if (*src == c)
	    *dest = *src++;
	else
	    *dest++ = *src++;
    }
    *dest = '\0';
}

/*
  Return the DODS Root directory pathname.
*/

static char *
dods_root()
{
    static char *dods_root = 0;

    if (!dods_root)
	dods_root = (getenv("DODS_ROOT") ? getenv("DODS_ROOT") : DODS_ROOT);

    return dods_root;
}

/* This is broken. Matlab 4 under sol; won't link properly. Also, a silly way
   to find the version number. 11/24/98 jhrg */

/** 
    @return The major version number. */

static int
major_version()
{
    Matrix *pm = mexGetMatrix("version");
    char p[MAX_STR], *r;
    
    if (!pm) {
	DBG(fprintf(stderr, "Could not read the version number\n"));
	return 0;
    }

    mxGetString(pm, &p[0], MAX_STR-1);

    r = strtok(p, ".");
    DBG(fprintf(stderr, "version: %s, %s\n", r, p));
    if (r)
	return atoi(r);
    else
	return 0;
}

/*
  Escape all characters that the Bourne shell treats as special. 
*/

static char *
quote_for_shell(char *src)
{
    char *c;
    char *dest = mxCalloc(2 * strlen(src), sizeof(char));
    
    strcpy(dest, "");
    while ((c = strpbrk(src, "\"*?[];&()|^<>"))) {
	strncat(dest, src, c - src);
	strcat(dest, "\\");
	strncat(dest, c, 1);
	c++;
	src = c;
    }
    strcat(dest, src);
    
    return dest;
}
			  
/** Probe for an executable image. Check to see if an executable copy of
    #pgm# resides in any of the paths in the the colon-separated list of
    paths #path_list#. 

    NB: The returned pathname is held in static storage. Subsequent calls
    will overwrite the values returned by previous calls.

    @return The first pathname found where #pgm# resides or NULL.
    @param path_list The colon-separated list of paths.
    @param pgm The program to search for. */


char *
probe_path(char *path_list, char *pgm)
{
    static char pathname[MAX_STR]; /* Bad */

    char *current = strtok(path_list, ":");
    while (current) {
	/* look for components that begin with `~' */
	if (current[0] == '~') {
	    strcpy(pathname, (getenv("HOME") ? getenv("HOME") : ""));
	    strcat(pathname, &current[1]);
	}
	else {
	    strcpy(pathname, current);
	}

	strcat(pathname, "/");
	strcat(pathname, pgm);

	DBG(msg("Testing pathname: %s\n", pathname));
	if (access(pathname, X_OK) == 0)
	    return pathname;

	current = strtok(NULL, ":");
    }

    return NULL;
}

/** Search the PATH, MATLABPATH and DODS_ROOT environment variables for a
    given executable. The pathname is returned in static storage.

    @return The full pathname of the executable, in static storage. 
    @param pgm The program to search for.
    @see probe_path() */

static char *
get_pathname(char *pgm)
{
    static char *path = NULL;
    static char *matlabpath = NULL;
    static char *dods_root_path = NULL;
    static char *pathname;
    
    /* Allocate this once. */
    if (!path)
	path = (char *)malloc(strlen(getenv("PATH") + 1));

    strcpy(path, getenv("PATH"));
    pathname = probe_path(path, pgm);
    if (pathname)
	return pathname;

    /* Allocate this once, too. */
    if (!matlabpath)
	matlabpath = (char *)malloc(strlen(getenv("MATLABPATH") + 1));

    strcpy(matlabpath, getenv("MATLABPATH"));
    pathname = probe_path(matlabpath, pgm);
    if (pathname)
	return pathname;

    /* If the program is in the CWD, use that copy. */
    pathname = probe_path(".", pgm);
    if (pathname)
	return pathname;

    /* Allocate this once, ... */
    if (!dods_root_path)
	dods_root_path = (char *)malloc(strlen(dods_root()) + 5);
    
    sprintf(dods_root_path, "%s/etc", dods_root());
    pathname = probe_path(dods_root_path, pgm);
    if (pathname)
	return pathname;

    sprintf(dods_root_path, "%s/bin", dods_root());
    pathname = probe_path(dods_root_path, pgm);
    if (pathname)
	return pathname;

    return NULL;
}

/* 
  Given a type name, return a pointer to the reader function (which is
  named do_* in this source file). Returns NULL if there is no matching
  function for TYPENAME.
*/

reader_type
get_reader(char *typename)
{
    int k = 0;
    do {
	if (strcmp(typelist[k].typename, typename) == 0)
	    return typelist[k].do_type;
    } while (typelist[++k].typename[0]);

    return NULL;
}

/*
  Read a line from writeval which should have only one word on it. Store the
  word in NAME. 

  NB: if this function is called to read in the name of a component in an
  aggregate type *and* we are prefixing names using the aggregate names then
  NAME must contain that prefix string. If *not* be *sure* to set the first
  element of NAME to '\0' before calling this function.
*/

static int
read_name(FILE *fin, char name[MAX_STR], int verbose)
{
    int items;
    char line[256];

    /* Pick up the next line that does not start with a newline character. */
    while (!feof(fin)) {
	if (!fgets(line, 255, fin)) {
	    if (errno == EINTR)
		continue;
	    if (verbose)
		err_msg("read_name: Could not read variable name or datatype.\n");
	    return FALSE;
	}
	if (line[0] != '\n')
	    break;
    }

    DBG(msg("read_name: line: `%s'\n", line));

    items = sscanf(line, "%s", &name[strlen(name)]);
    if (items != 1) {
	if (verbose)
	    err_msg("read_name: Could not parse variable name or datatype.\n");
	return FALSE;
    }

    return TRUE;
}

/* 
   Note: when reading an array description, `length' is the number of
   dimensions in the array.

   NB: See note above about NAME.
*/

static int
read_vec_descript(FILE *fin, char element_type[MAX_STR], char name[MAX_STR], 
		  int *length)
{
    int items;
    char line[256];

    if (feof(fin)) {
	DBG(msg("EOF found.\n"));
	return FALSE;
    }

    fgets(line, 255, fin);

    DBG(msg("line: `%s'\n", line));

    items = sscanf(line, "%s %s %d", &element_type[0], 
		   &name[strlen(name)], length);
    if (items !=3) {
	err_msg("read_vec_descript: Could not read type, name and/or length of vector.\n");
	return FALSE;
    }
    
    return TRUE;
}

/*
  Note that this code used to use the number of dimensions but no longer
  does. 
*/

static int
read_array_dims(FILE *fin, int dims[2])
{
    int items;
    int i = 0;
    char line[256];
    char *dimension;

    if (feof(fin)) {
	DBG(msg("EOF found.\n"));
	return FALSE;
    }

    fgets(line, 255, fin);

    DBG(msg("line: `%s'\n", line));

    /* Read in the dimension sizes and compute total number of elements.
       NB: Look for dimensions with size equal to one. These are dimensions
       that have been `nulled' by a constraint expression (reducung the
       dimensionality of an array from, say, three to two. Make sure that the
       first two elements of DIMS contain the valid sizes and not the size of
       the `nulled' dimenstion (which is one by definition). */
    dimension = strtok(line, " \n");
    while (dimension) {
	int dim = atoi(dimension);
	if (dim > 1) {
	    if (version_major < 5 && i >= 2) {
		err_msg("read_vec_descript: Matlab 4 does not support variables with more than two dimensions.\n");
		return FALSE;
	    }
	    dims[i++] = dim;
	}
	dimension = strtok(NULL, " \n");
    }

    return TRUE;
}

/*
  Read a name and length from writeval.

  NB: See note above about NAME.
*/

static int
read_aggregate(FILE *fin, char name[MAX_STR], int *length)
{
    int items;
    char line[256];
    fgets(line, 255, fin);

    DBG(msg("line: `%s'\n", line));

    items = sscanf(line, "%s %d", &name[strlen(name)], length);
    if (items != 2) {
	err_msg("read_aggregate: Could not read name and/or length of aggregate.\n");
	return FALSE;
    }
    
    return TRUE;
}

/*
  Read a line and discard it. This function is intended to be used to read
  the newline that follows a block of data. 

  NB: I used fseek() for this but that caused failuers on SunOS (and maybe
  Solaris?) for transfers that were an integral multiple of 2 (e.g., 32^2,
  512^2, 1023^2).
*/

static int
read_trailing_newline(FILE *fin)
{
    char dummy[256];
    if (feof(fin)) {
	err_msg("read_trailing_newline: Unexpected EOF.\n");
	return FALSE;
    }

    return fgets(dummy, 255, fin) != 0;
}

/* Build a vector containing the names of all the variables (as they appear
   in the datasets) for each variable read from the DODS server. Each name is
   separated by SEPARATOR_CHAR. */

/* NB: SPEARATOR_CHAR *must* be a string literal, not a character. */
#define SEPARATOR_CHAR "\t"

static void
build_variable_name_vector()
{
    char *name;
    char *names;
    int total_len = 0;

    /* Compute length of names vector. */
    name = get_variable_name(TRUE);
    while (name) {
	total_len += strlen(name);
	total_len++;		/* Add space for separator */
	name = get_variable_name(FALSE);
    }

    names = mxCalloc(total_len, sizeof(char));

    /* Load strings into names vector. */ 
    name = get_variable_name(TRUE);
    names[0] = '\0';
    while (name) {
	strcat(names, name);
	strcat(names, SEPARATOR_CHAR);
	name = get_variable_name(FALSE);
    }

    /* When this function is called it must have been the case that the first
       argument was put in `return_args[1]' *NOT* return_args[0]. However, if
       the -s (return names in string vector) option is not given, never call
       this function and use return_args[0], ... for the values. */

    return_args[0] = mxCreateString(names);
}

/* 
   This function never returns (but has return type int for compatibility
   with the other do_* functions.
*/

int 
do_error(FILE *fin, variable **vectors, char *prefix, int outermost)
{
    char value[BIG_STR];
    char val2[BIG_STR];
    int i = 0;

    value[0]='\0';
    fgets(value, BIG_STR-1, fin);
    do {
	i += strlen(&value[i]);
	fgets(&value[i], BIG_STR-(i+1), fin);
    } while (!feof(fin) && i < BIG_STR && value[i] != '\0');
	
    strip(val2, value, '"');		/* Strip out double quotes. */
    
    err_msg("writeval: %s\n", val2);
}

int
do_float(FILE *fin, variable **vectors, char *prefix, int outermost)
{
    double y;
    char name[MAX_STR] = "";
    variable *var;

    DBG2(msg("do_float: prefix: %s\n", prefix));

    /* If prefix is non-null load it into NAME. */
    if (fullnames && prefix[0]) {
	strcpy(name, prefix);
	strcat(name, "_");
    }
    else {
	name[0]='\0';
    }

    DBG2(msg("do_float: prefix: %s, name: %s\n", prefix, name));

    if (!read_name(fin, name, TRUE))
	return FALSE;

    DBG2(msg("do_float: name: %s\n", name));

    /* Check to see if this vector already exists. If not, create it,
       reassign the global list of vectors and set the local pointer VAR to
       the newly added element. */
    var = find_var(*vectors, name);
    if (!var)
	var = *vectors = add_var(*vectors, name, float64);

    /* Add the value. */
    fread(&y, sizeof(double), 1, fin);
    if (isnan(y)) {
	DBG(msg("Found NaN!\n"));
#if TEST_LOADDODS
	y = MakeNaN();
#else
	y = mexGetNaN();
#endif
    }
    add_float64_value(var, y);
    
    DBG(msg("Float value read: %lf\n", y));

#ifdef DODS_DEBUG2
    {
	union {
	    unsigned int ul[2];
	    double d;
	} v;
	v.d = y;
	msg("y (hex): %x %x\n", v.ul[0], v.ul[1]);
    }
#endif

    DBG2(msg("do_float: read: %s = %f\n", name, y));

    return TRUE;
}

int 
do_string(FILE *fin, variable **vectors, char *prefix, int outermost)
{
    variable *var;
    char value[BIG_STR];
    char name[MAX_STR] = "";

    /* If prefix is non-null load it into NAME. */
    if (fullnames && prefix[0]) {
	strcpy(name, prefix);
	strcat(name, "_");
    }
    else {
	name[0]='\0';
    }

    if (!read_name(fin, name, TRUE))
	return FALSE;

    /* Check to see if this vector already exists. If not, create it,
       reassign the global list of vectors and set the local pointer VAR to
       the newly added element. */
    var = find_var(*vectors, name);
    if (!var)
	var = *vectors = add_var(*vectors, name, string);

    /* Read and add the value; also read the trailing newline. */
    /* NB: This code used read_name() here, however, that introduced an error
       when reading attributes because they frequently contain spaces.
       read_name() would truncate the string at the first space thus cutting
       short the attribute value. */
    value[0]='\0';
    fgets(value, BIG_STR-1, fin);
    /* Strip ending new line if it exists. */
    if (value[strlen(value)-1] == '\n')
	value[strlen(value)-1] = '\0';

    add_string_value(var, &value[0]);

    DBG(msg("do_string: read: %s = %s\n", name, value));

    return TRUE;
}

/*
  Read the List variable into a single column vector.

  Untested.
*/

int
do_list(FILE *fin, variable **vectors, char *prefix, int outermost)
{
    int length;
    char element_type[MAX_STR];
    char name[MAX_STR] = "";	/* PREFIX + short_name */

    /* If prefix is non-null load it into NAME. */
    if (fullnames && prefix[0]) {
	strcpy(name, prefix);
	strcat(name, "_");
    }
    else {
	name[0]='\0';
    }

    if (!read_vec_descript(fin, element_type, name, &length))
	return FALSE;

    /* Process arrays of simple numeric types specially. */
    if (strcmp(element_type, "Byte") == 0
	|| strcmp(element_type, "Int32") == 0
	|| strcmp(element_type, "Float64") == 0) {
	double *yp = (double *) mxCalloc(length, sizeof(double));

	if (!yp) 
	    return FALSE;

	if (fread(yp, sizeof(double), length, fin) != length) {
	    err_msg("do_list: Could not read all the values.\n");
	    return FALSE;
	}

	if (!intern(name, length, 1, yp, extend_existing_variables))
	    return FALSE;

	if (verbose && !num_return_args)
	    msg("Creating vector %s with %d elements.\n", name, length);
    }
    else {
	int i;
	int status;
	reader_type exectype = get_reader(element_type);

	if (!exectype) {
	    err_msg("do_list: Error in dataset: Unknown datatype.\n");
	    return FALSE;
	}

	DBG2(msg("do_list: Processing: %s\n", element_type)); 

	for (i = 0; i < length; ++i) {
	    status = (*exectype)(fin, vectors, name, outermost);
	    if (!status) {
		err_msg("do_list: Error reading data stream.\n");
		return FALSE;
	    }
	}
    }

    if (!read_trailing_newline(fin))
	return FALSE;

    return TRUE;
}

/* Rudimentary testing */

/* Matlab 4 only supports 0,1,2 dims */
#define MAX_DIMS 2

int
do_array(FILE *fin, variable **vectors, char *prefix, int outermost)
{
    int i;
    int num_dims;
    int num_elems = 1;
    int dims[MAX_DIMS] = {1, 1}; /* Matlab 4 only supports 0,1,2 dims */
    int valid_dims = 0;
    char element_type[MAX_STR];
    char name[MAX_STR] = "";	/* PREFIX + short_name */

    /* If prefix is non-null load it into NAME. */
    if (fullnames && prefix[0]) {
	strcpy(name, prefix);
	strcat(name, "_");
    }
    else {
	name[0]='\0';
    }

    if (!read_vec_descript(fin, element_type, name, &num_dims))
	return FALSE;

    if (!read_array_dims(fin, dims))
	return FALSE;

    for (i = 0; i < MAX_DIMS; i++)
	if (dims[i] > 1) {
	    valid_dims++;
	    num_elems *= dims[i];
	}

    /* Process arrays of simple numeric types specially. */
    if (strcmp(element_type, "Byte") == 0
	|| strcmp(element_type, "Int16") == 0
	|| strcmp(element_type, "UInt16") == 0
	|| strcmp(element_type, "Int32") == 0
	|| strcmp(element_type, "UInt32") == 0
	|| strcmp(element_type, "Float32") == 0
	|| strcmp(element_type, "Float64") == 0) {
	double *yp = (double *) mxCalloc(num_elems, sizeof(double));

	if (!yp) {
	    err_msg("do_array: Could not allocate data array (%d bytes)\n",
		    num_elems);
	    return FALSE;
	}
	
	fread(yp, sizeof(double), num_elems, fin);

	/* Make sure to intern only column-major data! jhrg 3/27/97 */
	if (MAX(dims[0], 1) > 1 && MAX(dims[1], 1) > 1) {
	    double *ypcm = mxCalloc(num_elems, sizeof(double));

	    rm2cm(ypcm, yp, MAX(dims[0], 1), MAX(dims[1], 1));

	    mxFree(yp);
	    yp = ypcm;
	}

	/* NB: `dims' must have at least two elements. */

	if (!intern(name, MAX(dims[0], 1), MAX(dims[1], 1), yp, 
		    extend_existing_variables))
	    return FALSE;

	if (verbose && !num_return_args)
	    switch (valid_dims) {
	      case 0:
		msg("Creating scalar %s.\n", name);
		break;
	      case 1:
		msg("Creating vector %s with %d elements.\n",
		    name, num_elems);
		break;
	      case 2:	
		msg("Creating matrix %s (%d by %d) with %d elements.\n",
		    name, dims[0], dims[1], num_elems);
		break;
	      default:
		break;
	    }
    }
    else {
	int status;
	reader_type exectype = get_reader(element_type);

	if (!exectype) {
	    err_msg("do_array: Error in dataset: Unknown datatype.\n");
	    return FALSE;
	}

	DBG2(msg("do_array: Processing: %s\n", element_type)); 

	for (i = 0; i < num_elems; ++i) {
	    status = (*exectype)(fin, vectors, name, outermost);
	    if (!status) {
		err_msg("do_array: Error reading data stream.\n");
		return FALSE;
	    }
	}
    }

    if (!read_trailing_newline(fin))
	return FALSE;

    return TRUE;
}

/* Untested */

int
do_structure(FILE *fin, variable **vectors, char *prefix, int outermost)
{
    int num_elems;
    int i;
    char name[MAX_STR] = "";	/* PREFIX + short_name */

    /* If prefix is non-null load it into NAME. */
    if (fullnames && prefix[0]) {
	strcpy(name, prefix);
	strcat(name, "_");
    }
    else {
	name[0]='\0';
    }

    if (!read_aggregate(fin, name, &num_elems))
	return FALSE;

    DBG2(msg("do_struct: name: %s, elems: %d\n", name, num_elems));

    for (i = 0; i < num_elems; ++i) {
	int status;
	char  element_type[MAX_STR];
	reader_type exectype;

	element_type[0]='\0';
	if (!read_name(fin, element_type, TRUE))
	    return FALSE;

	DBG2(msg("do_struct: Processing: %s\n", element_type)); 

	exectype = get_reader(element_type);

	if (!exectype) {
	    err_msg("do_struct: Error in dataset: Unknown datatype.\n");
	    return FALSE;
	}

	status = (*exectype)(fin, vectors, name, outermost);
	if (!status) {
	    err_msg("do_struct: Error reading data stream.\n");
	    return FALSE;
	}
    }
    
    return TRUE;
}

int
do_sequence(FILE *fin, variable **vectors, char *prefix, int outermost)
{
    int num_elems;
    int i;
    char name[MAX_STR] = "";	/* PREFIX + short_name */

    /* If prefix is non-null load it into NAME. For Seqs, don't do this on
       the outermost level to conserve name length. */
    if (fullnames && prefix[0] && !outermost) {
	strcpy(name, prefix);
	strcat(name, "_");
    }
    else {
	name[0]='\0';
    }

    if (!read_aggregate(fin, name, &num_elems))
	return FALSE;

    DBG(msg("do_seq: name: %s, elems: %d\n", name, num_elems));

    /* NB: Run the loop util all the data is used up when reading the
       outermost sequence, otherwise read the elements only once. */
    do {
	for (i = 0; i < num_elems; ++i) {
	    int status;
	    char  element_type[MAX_STR];
	    reader_type exectype;

	    element_type[0]='\0';
	    DBG(msg("element type: %s\n", element_type));
	    if (!read_name(fin, &element_type[0], FALSE)) {
		if (outermost)
		    return TRUE;
		else {
		    err_msg("do_sequence: Could not read element type.\n");
		    return FALSE;
		}
	    }

	    DBG(msg("do_seq: Processing: %s\n", element_type)); 

	    exectype = get_reader(element_type);
	    if (!exectype) {
		err_msg("do_sequence: Error in dataset: Unknown datatype (%s).\n",
			element_type);
		return FALSE;
	    }

	    status = (*exectype)(fin, vectors, name, FALSE);
	    if (!status) {
		err_msg("do_sequence: Error reading data stream.\n");
		return FALSE;
	    }
	}
    } while (outermost && !feof(fin) && !ferror(fin));

    return TRUE;
}

int
do_grid(FILE *fin, variable **vectors, char *prefix, int outermost)
{
    char name[MAX_STR] = "";	/* PREFIX + short_name */
    int num_elems;

    /* If prefix is non-null load it into NAME. */
    if (fullnames && prefix[0]) {
	strcpy(name, prefix);
	strcat(name, "_");
    }
    else {
	name[0]='\0';
    }

    if (!read_name(fin, name, TRUE))
	return FALSE;

    DBG(msg("do_grid: name: %s\n", name));

    {
	int i;
	int status;
	char part[MAX_STR];
	for (i = 0; i < 2; ++i) {
	    int j;

	    part[0]='\0';
	    if (!read_aggregate(fin, part, &num_elems))
		return FALSE;

	    DBG(msg("do_grid: part: %s, num_elems: %d\n", part, num_elems));

	    for (j = 0; j < num_elems; ++j) {
		char dummy[MAX_STR];

		DBG2(msg("do_grid: Processing: Array\n")); 

		dummy[0]='\0';
		if (!read_name(fin, dummy, TRUE))
		    return FALSE;

		status = do_array(fin, vectors, name, FALSE);
		if (!status) {
		    err_msg("do_grid: Error reading data stream.\n");
		    return FALSE;
		}
	    }
	}
    }

    return TRUE;
}

/* 
  Read the command arguments and perform initial actions. If the command is
  called with no argument or with a "*" or "?", start up the WWW browser and
  open the default DODS locator page. If the argument is "*" or "?" followed
  by a URL then use that URL as the `locator page'. Finally, if the command
  is given a URL (without a prefix of "*" or "?") return that without any
  further actions. If the command argument ends with the suffix ".mat",
  assume that it is a matlab `mat' file and load it without any additional
  processing.
  
  NB: Memory allocated by mxCalloc() is automatically freed on exit from the
  command (by Matlab), so there is no need to call mxFree() on mxCalloc'd
  blocks. 

  Returns: A URL to a DODS dataset or NULL if no further action is required. 
*/

#ifndef TEST_LOADDODS

char *
init(int nlhs, Matrix *plhs[], int nrhs, Matrix *prhs[])
{
    char *s = mxCalloc(MAX_QUEUE_LEN, sizeof(char)); /* Return this array. */
    int i;			/* Count prhs[] elements. */

    if (!s)
	return NULL;

    /* Broken. See comments where major_verison() is defined. */
    version_major = major_version();

    for (i = 0; i < nrhs; ++i)
	if (!mxIsString(prhs[i])) {
	    msg("usage: arguments must be strings.\n");
	    return NULL;
	}

    /* If there are two arguments, assume that the first is a set of command
       line options. Extract them and set the global `command_opt' string. */

    if (nrhs == 0) {
	command_opts[0] = '\0';
	s[0] = '\0';
    }
    else if (nrhs == 1) {
	command_opts[0] = '\0';
	mxGetString(prhs[0], s, (MAX_QUEUE_LEN-1));
	/* If the fisrt char of `s' is `-' assume that s contains options 
           and since nrhs == 1, that no URL or filename was given. */
	if (s[0] == '-') {
	    strncpy(command_opts, s, (MAX_STR-1));
	    command_opts[(MAX_STR-1)] = '\0';
	    s[0] = '\0';
	}
    }
    else if (nrhs == 2) {
	mxGetString(prhs[0], command_opts, (MAX_STR-1));
	mxGetString(prhs[1], s, (MAX_QUEUE_LEN-1));
    }	
    else {			/* (nrhs > 2 || nrhs < 1) */
	msg("usage: loaddods [options] [url | `*' | `?']\n");
	return NULL;
    }

    /* Scan the command options string and set various flags. */

    /* Check for the `-e' option and call err_init(). */
    {
    	char *err_opt;
	if ((err_opt = strstr(command_opts, "-e"))) {
	    err_init(0);	/* user asks for new scheme */
	    /* blank out `-e' since writeval does not understannd it */
	    (*err_opt) = ' '; *(err_opt + 1) = ' '; 
	}
	else 
	    err_init(1);	/* else, use compatibility mode */
    }

    /* Print the version of both the loaddods and writeval programs. */
    if (strstr(command_opts, "-V")) {
	FILE *fin;
	char *pathname = get_pathname(DODS_BASE_COMMAND);
	char *command = mxCalloc(strlen(command_opts) + strlen(pathname)
				 + strlen(DODS_COMPLETE_COMMAND) + 1, 
				 sizeof(char));
	/* Add an error message here. */
	if (!pathname) {
	    err_msg("init: Could not find writeval.\n");
	    return NULL;
	}

	/* If the user called loaddods('-V') using the assignment syntax,
	   then make user that the version number string is assigned to the
	   LHS identifier given by the user. Otherwise, dump the version
	   strings to stderr. 4/20/99 jhrg */
	if (nlhs > 0) {
	    int error = 0;
	    char ver[MAX_STR];
	    /* Route stderr (which is where writeval sends the version
	       information) to stdout so that we can read it from pin. */
	    sprintf(command, DODS_COMPLETE_COMMAND, pathname, 
		    command_opts, "2>&1");
	    DBG(msg("command: %s\n", command));
	    fin = popen(command, "r");
	    fgets(ver, MAX_STR, fin);
	    DBG(msg("Writeval version string: %s\n", ver));
	    pclose(fin);
	    strcat(ver, "loaddods: ");
	    strcat(ver, dods_version);
	    plhs[0] = mxCreateString(ver);
	    mxSetName(plhs[0], "dods_vesion");
	}
	else {
	    sprintf(command, DODS_COMPLETE_COMMAND, pathname, 
		    command_opts, "");
	    DBG(msg("command: %s\n", command));
	    fin = popen(command, "r");
	    pclose(fin);
	    msg("loaddods: %s\n", dods_version);
	}

	return NULL;
    }

    /* Check for `+v' anti-verbose option. If not present, use verbose mode
       of writeval (since -v is the default for loaddods). */
    {
	char *no_verb;
	if ((no_verb = strstr(command_opts, "+v"))) {
	    verbose = 0;
	    /* blank out `+v' since writeval does not understannd it */
	    (*no_verb) = ' '; *(no_verb + 1) = ' '; 
	}
	else if (!strstr(command_opts, "-v"))
	    strcat(command_opts, " -v");
    }

    /* Check for `-k' catenate option. If present, turn on variable
       catenation by setting the global extend_existing_variables. Variable
       catenation is off by default. 

       Note that writeval is clueless about this option. */
    {
	char *kat;
	if ((kat = strstr(command_opts, "-k"))) {
	    extend_existing_variables = 1;
	    /* blank out `-k' since writeval does not understannd it */
	    (*kat) = ' '; *(kat + 1) = ' '; 
	}
	else
	    extend_existing_variables = 0;
    }

    /* Check for the `fullnames' option. If present, name each variable using
       all the names of the constructor types that contain the variable.
       Otherwise, name variables using just the leaf variable's name. NB:
       Since Matlab 4 limits variable names to 31 characters it is easy to
       create names that are too long using this option. It is better to use
       writeval's -r option to rename variables. */
    {
	char *full;
	if ((full = strstr(command_opts, "-n"))) {
	    fullnames = 1;
	    /* blank out `-n' since writeval does not understannd it */
	    (*full) = ' '; *(full + 1) = ' '; 
	}
	else
	  fullnames = 0;
    }

    /* Did the user supply `-s'? If so, output a vector/matrix containing the
       names of each of the variables returned via the `return args' feature.
       Note that this has no effect when interning variables directory into
       the Matlab workspace. */
    {
	char *names;
	if ((names = strstr(command_opts, "-s"))) {
	    if (nlhs > 0) {	/* only return values if they are there! */
		output_variable_names = 1;
	    }
	    else {
		msg("The -s option (return a vector of variable names) is\n");
		msg("is only available when using loaddods with return\n");
		msg("arguments. See the online help for more information.\n");
	    }
	    /* blank out `-s' since writeval does not understannd it */
	    (*names) = ' '; *(names + 1) = ' '; 
	}
	else
	    output_variable_names = 0;
    }

    /* Check for the short-names option. When this is given all names longer
       than 19 characters are truncated to 19 chars for compatibility with
       Matlab 4. This option is is signalled by -T. */

    {
	char *trunc;
	if ((trunc = strstr(command_opts, "-T"))) {
	    *trunc = ' ';	/* Blank out -T since writeval */
	    *(trunc + 1) = ' '; /* does not understand it. */
	    truncate_variable_names = TRUE;
	}
	else
	    truncate_variable_names = FALSE;
    }
	    

    /* END of option processing code */

    /* Now deal with the return arguments. If nlhs > 0 use plhs[] for the
       return values, otherwise directly intern values. */
    if (nlhs > 0) {
	num_return_args = nlhs;
	return_args = &plhs[0];
	if (output_variable_names)
	    current_arg = 1;	/* First return arg is string matrix. */
	else
	    current_arg = 0;
	if (extend_existing_variables) {
	    msg("The option `-k' (extend existing variables) has no effect\n");
	    msg("when using loaddods with a vector of return variables\n");
	    extend_existing_variables = 0;
	}
    }
    else {
	num_return_args = 0;
    }

    if (s[0] == '*' || s[0] == '?' || s[0] == '\0') {
	key_t msgqid = 3075;
	long queue_id;

	if (strlen(s) == 1)
	    strcpy(s, DEFAULT_LOCATOR);

	if (strlen(s) > 1) {
	    char *command = mxCalloc(strlen(s) + strlen(BROWSER_COMMAND) + 1,
				     sizeof(char));
	    if (!command)
		return NULL;

	    sprintf(command, BROWSER_COMMAND, s);
	    DBG(msg("Command: %s\n", command));
	    system(command);
	    mxFree(command);
	}

	/* Read the URL from the WWW browser (using a SYS V queue). */

	if (questart(&msgqid) < 0)
	    err_msg("Queue exists already.");

	queue_id = queconn(&msgqid);
	msg("Waiting for URL from browser...\n");

	if (queread(queue_id, s, 10) < 0) {
	    err_msg("init: Cannot read from queue.\n");
	    return NULL;
	}

	questop(queue_id);
    }

    return s;
}

void 
mexFunction(int nlhs, Matrix *plhs[], int nrhs, Matrix *prhs[])
{
    int childpid;
    char *command;
    char *dods_url;
    char *pathname;
    char datatype[MAX_STR];	/* Type of variable. */
    variable *vectors = NULL;
    FILE *fin, *fout;		/* fout is unused. */

    clear_existing_variables();	/* this is for the array catenation code */

    if (!(dods_url = init(nlhs, plhs, nrhs, prhs)))
	return;

    dods_url = quote_for_shell(dods_url);

    pathname = get_pathname(DODS_BASE_COMMAND);
    if (!pathname) {
	err_msg("mexFunction: Could not locate the `writeval' helper application.\n");
	return;
    }

    command = mxCalloc(strlen(pathname) +strlen(dods_url) 
		       + strlen(command_opts) + strlen(DODS_COMPLETE_COMMAND)
		       + 1, sizeof(char));

    if (!command)
	return;

    sprintf(command, DODS_COMPLETE_COMMAND, pathname, command_opts, dods_url);
    DBG(msg("command: %s\n", command));
    fin = popen(command, "r");

    if (!fin || feof(fin) || ferror(fin)) {
	err_msg("mexFunction: Could not link with DODS client subprocess\n");
	return;
    }

    mxFree(command);

    /* Note that below read_name() is called with VERBOSE == FALSE since this
       call will fail when there are no more variables to read. */

    datatype[0]='\0';
    while (read_name(fin, &datatype[0], FALSE) && !ferror(fin) && !feof(fin)) {
	int status;		/* Return value from exectype(). */
	reader_type exectype;

	DBG2(msg("Processing: %s\n", datatype));

	exectype = get_reader(datatype);
	if (!exectype) {
	    err_msg("mexFunction: Error in dataset: Unknown datatype (%s).\n", datatype);
	    return;
	}

	status = (*exectype)(fin, &vectors, "", TRUE);
	if (!status) {
	    err_msg("mexFunction: Data reader: Error reading data stream.\n");
	    return;
	}

	fflush(stderr);
	fflush(stdout);
	datatype[0]='\0';
    }

    if (ferror(fin)) {
	err_msg("mexFunction: Subprocess IPC: Error reading data stream.\n");
	return;
    }

    /* Read any remaining characters in the pipe. */
    {
	char c;
	while (!feof(fin))
	    fread(&c, 1, 1, fin);
    }

    /* Close the pipe. */
    {
	int status = pclose(fin);
	if (status != 0)
	    err_msg("mexFunction: Error: DODS client subprocess: %d\n", status);
    }

    /* Now intern the vectors built up from ctor variable's multiple calls to
       do_float and do_string. Make each vector a column vector. */

    {
	variable *var = vectors;

	while (var) {
	    int status;
	    switch (var->type) {
	      case float64:
		status = intern(var->name, var->length, 1, var->data,
				 extend_existing_variables);
		if (status && verbose && !num_return_args) {
		    Matrix *mtmp = mexGetMatrixPtr(var->name);
		    
		    msg("Creating vector `%s' with %d elements.\n",
			var->name, mxGetM(mtmp));
		}
		break;

	      case string: {
		  char *chars = reorganize_strings(var, mxCalloc);
		  status = intern_strings(var->name, chars, 
					  extend_existing_variables);
		  if (status && verbose && !num_return_args)
		      msg("Creating string vector `%s'.\n", var->name);
		  break;
	      }

	      default:
		status = FALSE;
		break;
	    }

	    if (!status) {
		err_msg("mexFunction: Could not add vector `%s'.\n", var->name);
		return;
	    }

	    var = var->next;
	}
    }

    if (output_variable_names)
	build_variable_name_vector();

    fflush(stderr);
    fflush(stdout);

    free_vars(vectors);

    return;
}

#endif /* !TEST_LOADDODS */

#ifdef TEST_LOADDODS

/*
  The `test' version of loaddods takes a single argument: the dods url to
  fetch.
*/

int
main(int argc, char *argv[])
{
    int childpid;
    char *command;
    char *dods_url;
    char *pathname;
    char datatype[MAX_STR];	/* Type of variable. */
    variable *vectors = NULL;
    FILE *fin, *fout;		/* fout is unused. */

    if (argc != 2)
	return 1;

    dods_url = argv[1];
    dods_url = quote_for_shell(dods_url);

    pathname = get_pathname(DODS_BASE_COMMAND);
    if (!pathname) {
	msg("Could not locate the `writeval' helper application.\n");
	return;
    }

    command = mxCalloc(strlen(pathname) +strlen(dods_url) 
		       + strlen(command_opts) + strlen(DODS_COMPLETE_COMMAND)
		       + 1, sizeof(char));

    if (!command)
	return;

    /* Try directories along the Shell's path and the DODS_ROOT/etc. */
    sprintf(command, DODS_COMPLETE_COMMAND, pathname, command_opts, dods_url);
    DBG(msg("command: %s\n", command));
    fin = popen(command, "r");

    if (!fin || feof(fin) || ferror(fin)) {
	msg("Could not link with DODS client subprocess\n");
	return;
    }

    mxFree(command);

    datatype[0]='\0';
    while (read_name(fin, &datatype[0], FALSE) && !ferror(fin) && !feof(fin)) {
	int status;		/* Return value from exectype(). */
	reader_type exectype;

	DBG2(msg("Processing: %s\n", datatype));

	exectype = get_reader(datatype);
	if (!exectype) {
	    msg("Error in dataset: Unknown datatype.\n");
	    return 1;
	}

	status = (*exectype)(fin, &vectors, "", TRUE);
	if (!status) {
	    msg("Data reader: Error reading data stream.\n");
	    return 1;
	}

	fflush(stderr);
	fflush(stdout);
	datatype[0]='\0';
    }

    if (ferror(fin)) {
	msg("Subprocess IPC: Error reading data stream.\n");
	return 1;
    }

    {
	char c;
	while (!feof(fin))
	    fread(&c, 1, 1, fin);
    }


    {
	int status = pclose(fin);
	if (status != 0)
	    fprintf(stderr, "Error: DODS client subprocess: %d\n", status);
    }

    /* Now intern the vectors built up from ctor variable's multiple calls to
       do_float and do_string. Make each vector a column vector. */

    {
	variable *var = vectors;

	while (var) {
	    int status;
	    switch (var->type) {
	      case float64:
		msg("Float64 vector: %s, %d elements\n", var->name,
		    var->length);
		break;

	      case string: {
		  char *chars = reorganize_strings(var, calloc);
		  msg("Creating string vector `%s', %d elements\n",
		      var->name, var->length);
		  break;
	      }

	      default:
		status = FALSE;
		break;
	    }

	    if (!status) {
		msg("Could not add vector `%s'.\n", var->name);
		return 1;
	    }

	    var = var->next;
	}
    }

    fflush(stderr);
    fflush(stdout);

    free_vars(vectors);

    return 1;
}

#endif /* TEST_LOADDODS */
